from typing import List

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

from ..config.core import config


class DatetimeVariableEstimator(BaseEstimator, TransformerMixin):
    """Derived features from date"""

    def __init__(self, date_variable: str):

        if not isinstance(date_variable, str):
            raise ValueError("date_variable should be a string of format 'yyyy-mm-dd' ")

        self.date_variable = date_variable

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        # we need this step to fit the sklearn pipeline
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        # so that we do not over-write the original dataframe
        X = X.copy()
        X[self.date_variable] = pd.to_datetime(X[self.date_variable])

        # creating date based features
        X["year"] = X[self.date_variable].dt.year
        X["day_of_week_sin"] = np.sin(
            2
            * np.pi
            * X[self.date_variable].dt.dayofweek
            / max(X[self.date_variable].dt.dayofweek)
        )
        X["day_of_year_sin"] = np.sin(
            2
            * np.pi
            * X[self.date_variable].dt.dayofyear
            / max(X[self.date_variable].dt.dayofyear)
        )
        X["day_of_year_cos"] = np.cos(
            2
            * np.pi
            * X[self.date_variable].dt.dayofyear
            / max(X[self.date_variable].dt.dayofyear)
        )
        X["week"] = X[self.date_variable].dt.isocalendar().week
        X.set_index("date", inplace=True)

        return X


class StatisticalVariableEstimator(BaseEstimator, TransformerMixin):
    """Derived statistical features"""

    def __init__(self, prevision: str, effectif: str):

        if not isinstance(prevision, str):
            raise ValueError("prevision should be a string")

        if not isinstance(effectif, str):
            raise ValueError("effectif should be a string")

        # while concatenating X and y, y column gets rename 0
        self.reel = config.model_config.target
        self.prevision = prevision
        self.effectif = effectif

    def fit(self, X: pd.DataFrame, y: pd.Series):
        
        # to avoid data leakage we compute the relevant statistics on train set only
        y.index = X.index
        self.y = y

        X = X.copy()
        train = pd.concat((X, self.y), axis=1)

        # aggregate data significantly (canteen, scholar year and week level) and compute statistics
        train["freq_predicted_%"] = train[self.prevision] / train[self.effectif]
        train["freq_reel_%"] = train[self.reel] / train[self.effectif]
        grouped_train = train.groupby(["cantine_nom", "week", "annee_scolaire"])
        grouped_train_m = (
            grouped_train[["freq_predicted_%", "freq_reel_%", self.prevision, self.reel]]
            .mean()
            .reset_index()
        )
        grouped_train_s = (
            grouped_train[["freq_predicted_%", "freq_reel_%", self.prevision, self.reel]]
            .std()
            .reset_index()
        )

        # then we average on the years, to have a single number per canteen and week
        # in order to be able to spread it to the test set
        agg_mean = grouped_train_m.groupby(["cantine_nom", "week"])
        self.agg_mean = agg_mean[["freq_predicted_%", "freq_reel_%"]].mean()
        agg_std = grouped_train_s.groupby(["cantine_nom", "week"])
        self.agg_std = agg_std[["freq_predicted_%", "freq_reel_%"]].mean()

        self.agg_std.rename(
            columns={
                "freq_predicted_%": "freq_predicted_%_std",
                "freq_reel_%": "freq_reel_%_std",
            },
            inplace=True,
        )

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        X = X.merge(
            self.agg_mean, left_on=["cantine_nom", "week"], right_index=True, how="left"
        )

        X = X.merge(
            self.agg_std, left_on=["cantine_nom", "week"], right_index=True, how="left"
        )

        X.drop(["freq_predicted_%", "freq_predicted_%_std"], axis=1, inplace=True)
        X.reset_index(drop=True, inplace=True)

        return X


class NumericalImputer(BaseEstimator, TransformerMixin):
    """Numerical missing value imputer"""

    def __init__(self, variables: List[str]):

        if not isinstance(variables, list):
            raise ValueError("Variables should be in a list")

        self.variables = variables

    def fit(self, X, y: pd.Series = None):
        # persist median in a dictionary
        self.imputer_dict_ = {}

        for feature in self.variables:
            self.imputer_dict_[feature] = X[feature].median()
        return self

    def transform(self, X):

        X = X.copy()
        for feature in self.variables:
            X[feature].fillna(self.imputer_dict_[feature], inplace=True)
        return X


class TargetEncoder(BaseEstimator, TransformerMixin):
    """String to numbers categorical encoder."""

    def __init__(self, variables: List[str]):

        if not isinstance(variables, list):
            raise ValueError("Variables should be in a list")

        self.variables = variables

    def fit(self, X: pd.DataFrame, y: pd.Series):
        
        y.index = X.index
        self.y = y

        X = X.copy()
        
        temp = pd.concat((X, self.y), axis=1)
        temp.columns = list(X.columns) + ["target"]

        # persist enconding in dictionary
        self.encoder_dict_ = {}

        for var in self.variables:
            t = temp.groupby([var])["target"].mean().sort_values(ascending=True).index
            self.encoder_dict_[var] = {k: i for i, k in enumerate(t, 0)}

        return self

    def transform(self, X):

        X = X.copy()
        for feature in self.variables:
            X[feature] = X[feature].map(self.encoder_dict_[feature])
        return X



import logging

from attendance_model.config.core import PACKAGE_ROOT, config

# only add NullHandler to library’s loggers. Indeed the configuration
# of handlers is the prerogative of the application developer who uses your
# library. And knows their target audience.
logging.getLogger(config.app_config.package_name).addHandler(logging.NullHandler())


with open(PACKAGE_ROOT / "VERSION") as version_file:
    __version__ = version_file.read().strip()



import pandas as pd

from attendance_model.config.core import config
from attendance_model.pipeline import freq_pipeline
from attendance_model.processing.data_manager import load_dataset, save_pipeline


def timeseries_train_test_split(data, split_date):
    """
    Perform train-test split with respect to time series structure
    """

    split_date = pd.to_datetime(split_date)
    data["date"] = pd.to_datetime(data["date"])

    # create train test partition
    X_train = data.loc[data["date"] < split_date]
    X_test = data.loc[data["date"] >= split_date]

    y_train = X_train.reel
    X_train = X_train.drop(["reel"], axis=1)
    y_test = X_test.reel
    X_test = X_test.drop(["reel"], axis=1)

    return X_train, y_train, X_test, y_test


def zscore_outliers(data, column, n):
    """
    Tag outliers in dataset in regard from a specific column
    As an empirical rule, often 3 std from the mean is consider as an outlier
    Return two boolean columns : upper_outlier and lower_outlier
    """

    outliers = data[(data[column] != 0)]

    # we aggregate value by cantine and annee in order to have
    # a meaningfull outlier detection system and compare what's comparable
    outliers = outliers.groupby(["cantine_nom", "annee_scolaire"])
    outliers = outliers[column].agg(["mean", "std"])

    outliers["lower_bound"] = outliers["mean"] - (n * outliers["std"])
    outliers["upper_bound"] = outliers["mean"] + (n * outliers["std"])

    data = data.merge(
        outliers,
        left_on=["cantine_nom", "annee_scolaire"],
        right_index=True,
        how="left",
    )

    data["upper_outlier"] = data[column] > data["upper_bound"]
    data["lower_outlier"] = data[column] < data["lower_bound"]

    data.drop(["upper_bound", "lower_bound", "mean", "std"], axis=1, inplace=True)

    return data


def run_training() -> None:
    """Train the attendance model."""

    # read data
    data = load_dataset(file_name=config.app_config.data_file)

    # here i consider as an outlier any data point that sits
    # 2 std away from the mean, this rigourous choice is based on the fact
    # that Nantes metropole wants a model to predict "normal periods" first
    # most of the zeros are due to mistakes and are not meaningful values
    # strikes are not predictible by nature
    data = zscore_outliers(data, config.model_config.target, 2)
    data = data.loc[
        (data["greve"] != 1)
        & (data["reel"] != 0)
        & (data["upper_outlier"] != 1)
        & (data["lower_outlier"] != 1)
    ]
    data.drop(["greve", "upper_outlier", "lower_outlier"], axis=1, inplace=True)

    data = data.dropna(axis=0, subset=[config.model_config.target])

    # divide train and test
    X_train, y_train, X_test, y_test = timeseries_train_test_split(
        data, split_date=config.model_config.split_date
    )

    # transform target
    y_train = y_train ** (1 / 2)

    # fit model
    freq_pipeline.fit(X_train, y_train)

    print(f"Score: {freq_pipeline.score(X_train, y_train)}")

    # persist trained model
    save_pipeline(pipeline_to_persist=freq_pipeline)


if __name__ == "__main__":
    run_training()



from lightgbm import LGBMRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from attendance_model.config.core import config
from attendance_model.processing import feature_engineering as fe

freq_pipeline = Pipeline(
    [
        # ===== ADDITIONAL FEATURES =====
        (
            "date_features",
            fe.DatetimeVariableEstimator(date_variable="date"),
        ),
        (
            "statistical_features",
            fe.StatisticalVariableEstimator(prevision="prevision", effectif="effectif"),
        ),
        # ===== IMPUTATION =====
        # impute prevision variable with interpolate method
        (
            "median_imputation",
            fe.NumericalImputer(
                variables=config.model_config.numerical_vars_with_na,
            ),
        ),
        (
            "categ_imputation",
            fe.TargetEncoder(
                variables=config.model_config.categorical_vars_with_na,
            ),
        ),
        # ===== SCALING =====
        ("scaler", StandardScaler()),
        # ===== MODELING =====
        (
            "LGBR",
            LGBMRegressor(
                alpha=config.model_config.alpha,
                objective=config.model_config.objective,
                metric=config.model_config.metric,
                max_depth=config.model_config.max_depth,
                n_estimators=config.model_config.n_estimators,
                num_leaves=config.model_config.num_leaves,
                random_state=config.model_config.random_state,
            ),
        ),
    ]
)



import os
import sqlite3 as sql
import typing as t

import joblib
import pandas as pd
from sklearn.pipeline import Pipeline

from attendance_model import __version__ as _version
from attendance_model.config.core import DATASET_DIR, TRAINED_MODEL_DIR, config


def load_dataset(*, file_name: str) -> pd.DataFrame:
    """Load the necessary data from the datawarehouse."""

    conn = sql.connect(os.path.join(DATASET_DIR / file_name))

    SQL_Query = pd.read_sql_query(
        """select
    Frequentation_quotidienne.date,
    Frequentation_quotidienne.prevision,
    Frequentation_quotidienne.reel,
    Dim_site.cantine_nom,
    Dim_site.annee_scolaire,
    Dim_site.effectif,
    Dim_site.quartier_detail,
    Dim_site.prix_quartier_detail_m2_appart,
    Dim_site.prix_moyen_m2_appartement,
    Dim_site.prix_moyen_m2_maison,
    Dim_site.longitude,
    Dim_site.latitude,
    Dim_temporelle.depuis_vacances,
    Dim_temporelle.depuis_ferie,
    Dim_temporelle.depuis_juives,
    Dim_temporelle.ramadan_dans,
    Dim_temporelle.depuis_ramadan,
    Dim_evenement.greve

    from Frequentation_quotidienne

    left join Dim_site               on Frequentation_quotidienne.site_id = Dim_site.site_id
    left join Dim_menu               on Frequentation_quotidienne.jour_id = Dim_menu.jour_id
    left join Dim_temporelle         on Frequentation_quotidienne.jour_id = Dim_temporelle.jour_id
    left join Dim_evenement          on Frequentation_quotidienne.jour_id = Dim_evenement.jour_id

    order by Frequentation_quotidienne.jour_site_id
    """,
        conn,
    )

    dataframe = pd.DataFrame(
        SQL_Query,
        columns=[
            "date",
            "prevision",
            "reel",
            "cantine_nom",
            "annee_scolaire",
            "effectif",
            "quartier_detail",
            "prix_quartier_detail_m2_appart",
            "prix_moyen_m2_appartement",
            "prix_moyen_m2_maison",
            "longitude",
            "latitude",
            "depuis_vacances",
            "depuis_ferie",
            "depuis_juives",
            "ramadan_dans",
            "depuis_ramadan",
            "greve",
        ],
    )

    return dataframe


def save_pipeline(*, pipeline_to_persist: Pipeline) -> None:
    """
    Saves the versioned model, and overwrites any previous
    saved models.
    """

    # Prepare versioned save file name
    save_file_name = f"{config.app_config.pipeline_save_file}{_version}.pkl"
    save_path = TRAINED_MODEL_DIR / save_file_name

    remove_old_pipelines(files_to_keep=[save_file_name])
    joblib.dump(pipeline_to_persist, save_path)


def load_pipeline(*, file_name: str) -> Pipeline:
    """Load a persisted pipeline."""

    file_path = TRAINED_MODEL_DIR / file_name
    trained_model = joblib.load(filename=file_path)
    return trained_model


def remove_old_pipelines(*, files_to_keep: t.List[str]) -> None:
    """
    This is to ensure a simple one-to-one mapping between the package version
    and the model version to be imported and used by other applications.
    """
    do_not_delete = files_to_keep + ["__init__.py"]
    for model_file in TRAINED_MODEL_DIR.iterdir():
        if model_file.name not in do_not_delete:
            model_file.unlink()



from datetime import datetime
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from pydantic import BaseModel, ValidationError

from attendance_model.config.core import config


def drop_na_inputs(*, input_data: pd.DataFrame) -> pd.DataFrame:
    """
    Check model inputs for na values and filter them
    based on the ones the model knows how to handle.
    """

    validated_data = input_data.copy()
    new_vars_with_na = [
        var
        for var in config.model_config.inputs
        if var
        not in config.model_config.categorical_vars_with_na
        + config.model_config.numerical_vars_with_na
        and validated_data[var].isnull().sum() > 0
    ]
    validated_data.dropna(subset=new_vars_with_na, inplace=True)

    return validated_data


def validate_inputs(*, input_data: pd.DataFrame) -> Tuple[pd.DataFrame, Optional[dict]]:
    """Check model inputs for unprocessable values."""

    relevant_data = input_data[config.model_config.inputs].copy()
    validated_data = drop_na_inputs(input_data=relevant_data)
    errors = None

    try:
        # pydantic cannot validate inputs with np.nan, so we replace them with None
        MultipleDataInputs(
            inputs=validated_data.replace({np.nan: None}).to_dict(orient="records")
        )
        # catch validation error if try fails
    except ValidationError as error:
        errors = error.json()

    return validated_data, errors


class DataInputSchema(BaseModel):
    """It's possible to be more strict by removing the Optional key word.
    In that case, if we run the validation process and one column is missing
    or if one type isn't the correct one, pydantic will throw a validation error."""

    date: Optional[datetime]
    prevision: Optional[int]
    cantine_nom: Optional[str]
    annee_scolaire: Optional[str]
    effectif: Optional[int]
    quartier_detail: Optional[str]
    prix_quartier_detail_m2_appart: Optional[int]
    prix_moyen_m2_appartement: Optional[int]
    prix_moyen_m2_maison: Optional[int]
    longitude: Optional[float]
    latitude: Optional[float]
    depuis_vacances: Optional[int]
    depuis_ferie: Optional[int]
    depuis_juives: Optional[int]
    ramadan_dans: Optional[int]
    depuis_ramadan: Optional[int]


class MultipleDataInputs(BaseModel):
    """In case, mutliple inputs are provided."""

    inputs: List[DataInputSchema]



import typing as t

import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error, mean_absolute_error, median_absolute_error

from attendance_model import __version__ as _version
from attendance_model.config.core import config
from attendance_model.processing.data_manager import load_pipeline
from attendance_model.processing.validation import validate_inputs

pipeline_file_name = f"{config.app_config.pipeline_save_file}{_version}.pkl"
_freq_pipeline = load_pipeline(file_name=pipeline_file_name)


def make_prediction(
    *,
    input_data: t.Union[pd.DataFrame, dict],
) -> dict:
    """Make a prediction using a saved model pipeline."""

    data = pd.DataFrame(input_data) 
    if "reel" in data.columns:
        y_test = data.reel

    validated_data, errors = validate_inputs(input_data=data)
    results = {"predictions": None, "version": _version, "errors": errors}

    if not errors:

        predictions = _freq_pipeline.predict(X=validated_data)
        results = {
            "predictions": [pred ** 2 for pred in predictions],
            "version": _version,
            "errors": errors,
        }

    try:
        print(f"Root Mean Square Error: {str(np.sqrt(mean_squared_error(y_test, results['predictions'])))}")
        print(f"Mean Absolute Error: {str(mean_absolute_error(y_test, results['predictions']))}")
        print(f"Median Absolute Error: {str(median_absolute_error(y_test, results['predictions']))}")

    except:
        pass

    return results



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from pathlib import Path
from setuptools import find_packages, setup

# Package meta-data.
NAME = 'attendance-model'
DESCRIPTION = "An end to end machine learning app to predict canteens attendance 2-3 weeks ahead in Nantes Metropole"
URL = "https://github.com/qangelot/projet_Nantes"
EMAIL = "q.angelot@gmail.com"
AUTHOR = "qangelot"
REQUIRES_PYTHON = ">=3.6.0"


# ------------------------------------------------
# If changing the License, remember to change the
# Trove Classifier for that!
long_description = DESCRIPTION

# package's VERSION file in a dictionary
about = {}
ROOT_DIR = Path(__file__).resolve().parent
REQUIREMENTS_DIR = ROOT_DIR / 'requirements'
PACKAGE_DIR = ROOT_DIR / 'attendance_model'
with open(PACKAGE_DIR / "VERSION") as f:
   _version = f.read().strip()
   about["__version__"] = _version


# libraries required for module execution
def list_reqs(fname="requirements.txt"):
   with open(REQUIREMENTS_DIR / fname) as fd:
       return fd.read().splitlines()

# Where the magic happens:
setup(
   name=NAME,
   version=about["__version__"],
   description=DESCRIPTION,
   long_description=long_description,
   long_description_content_type="text/markdown",
   author=AUTHOR,
   author_email=EMAIL,
   python_requires=REQUIRES_PYTHON,
   url=URL,
   packages=find_packages(exclude=("tests",)),
   package_data={"attendance_model": ["VERSION"]},
   install_requires=list_reqs(),
   extras_require={},
   include_package_data=True,
   license="BSD-3",
   classifiers=[
       # Trove classifiers
       # standardized classifiers used by community members
       # to find projects based on their desired criteria
       "License :: OSI Approved :: MIT License",
       "Programming Language :: Python",
       "Programming Language :: Python :: 3",
       "Programming Language :: Python :: 3.6",
       "Programming Language :: Python :: 3.7",
       "Programming Language :: Python :: 3.8",
       "Programming Language :: Python :: 3.9",
       "Programming Language :: Python :: Implementation :: CPython",
       "Programming Language :: Python :: Implementation :: PyPy",
   ],
)